// ============================================================================
// $Id: intermud_comm.cc,v 1.3 1998/03/13 16:54:49 chris Exp $
//
// This is part of Realm Of Magic
//
// Usage: Intermud lowlevel network functionality wrapper class
//
// All rights reserved. See license.doc for complete information.
//
// Copyright (C) 1993-1996 by the Hackerz of the Realm Of Magic
// Realm Of Magic is based on CircleMUD, Copyright (C) 1993.
// CircleMUD is based on DikuMUD, Copyright (C) 1990, 1991.
// ============================================================================

#include "utility.hh"
#include "structs.hh"
#include "intermud_comm.hh"

_intermud::_intermud()
{
  bullet.bytes = 0;
}

void _intermud::Connect()
{
  if ((intermud_socket = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
    log("SYSERR: Failed to open InterMUD socket!");
    connection = false;
    return;
  }
  intermud_server.sun_family = AF_UNIX;
  strcpy(intermud_server.sun_path, MUDSOCK_PATH);
  adress_length = strlen(intermud_server.sun_path) + 
                   sizeof(intermud_server.sun_family);
  if (connect(intermud_socket, (struct sockaddr *) &intermud_server,
              adress_length) == -1) {
    log("WARNING: Intermud unavailable, couldn't connect!");
    connection = false;
    return;
  }
  log("Connection to Intermud successful!");
  connection = true;
  return;
}

bool _intermud::Connection()
{
  return connection;
}

string _intermud::ReceiveData()
{
  char raw_data[8192];
  int bytes;

  bytes = recv_data(raw_data);
  if (bytes <= 0) {
    mudlog("INTERMUD: Link dropped to Server - use connect!", NRM, LEVEL_GOD2,
           TRUE);
    close(intermud_socket);
    connection = false;
  } else
    raw_data[bytes] = '\0';

  in_data = raw_data;
  return in_data;
}

void _intermud::FlushData()
{
  int bytes;

  bytes = send_data();

  if (bytes <= 0) {
    log("SYSERR: Data send to MUD returned 0 bytes - dropping connection!");
    close(intermud_socket);
    connection = false;
    return;
  }

  return;
}

void _intermud::CloseFD()
{
  close(intermud_socket);
  connection = false;
  return;
}

string _intermud::OutBuffer()
{
  return out_data;
}

int _intermud::Desc()
{
  return intermud_socket;
}

int _intermud::recv_data(char *bf)
{
  int buflen, cc;

  cc = recv(intermud_socket, &bullet, sizeof(struct bullet), 0);
  if (cc <= 0)
    return cc;
  else {
    buflen = bullet.bytes;
    while (buflen > 0) {
      cc = recv(intermud_socket, bf, buflen, 0);
      if (cc <= 0) {
        log("SYSERR: _intermud::recv_data most probably fatal!");
        exit(1);
      }
      bf += cc;
      buflen -= cc;
    }
    return bullet.bytes;
  }
}

int _intermud::send_data()
{
  int buflen, cc;
  char *bf;
  
  bf = strdup(out_data.c_str());

  bullet.bytes = strlen(bf);
  
  if (bullet.bytes <= 0) {
    log("SYSERR: _intermud::send_data requested to send 0 bytes!");
    exit(1);
  }

  cc = send(intermud_socket, (char *) &bullet, sizeof(struct bullet), 0);
  if (cc <= 0) {
    free(bf);
    return cc;
  } else {
    buflen = bullet.bytes;
    while (buflen > 0) {
      cc = send(intermud_socket, bf, buflen, 0);
      if (cc <= 0) {
        log("SYSERR: _intermud::send_data probably fatal!");
      }
      bf += cc;
      buflen -= cc;
    }
    return bullet.bytes;
  }
}

_intermud& _intermud::operator<<(string s)
{
  this->out_data += s;
  return *this;
}

_intermud& _intermud::operator<<(int i)
{
  char num_buf[256];
  
  sprintf(num_buf, "%d", i);
  this->out_data += num_buf;
  return *this;
}

_intermud& _intermud::operator<<(unsigned int ui)
{
  char num_buf[256];

  sprintf(num_buf, "%u", ui);
  this->out_data += num_buf;
  return *this;
}

_intermud& _intermud::operator<<(long int li)
{
  char num_buf[256];

  sprintf(num_buf, "%ld", li);
  this->out_data += num_buf;
  return *this;
}

_intermud& _intermud::operator<<(unsigned long int uli)
{
  char num_buf[256];

  sprintf(num_buf, "%lu", uli);
  this->out_data += num_buf;
  return *this;
}

_intermud& _intermud::operator<<(short int si)
{
  return (*this) << (int)si;
}

_intermud& _intermud::operator<<(unsigned short int usi)
{
  return (*this) << (unsigned int)usi;
}

_intermud& _intermud::operator<<(_intermud& (*pf)(_intermud&))
{
  return (*pf)(*this);
}

_intermud& sendit(_intermud& flush_intermud)
{
  flush_intermud.FlushData();
  flush_intermud.out_data = "";
  return flush_intermud;
}

