// ============================================================================
// $Id: net.cc,v 1.1 1996/04/22 18:25:24 questor Exp $
//
// This is part of Realm Of Magic
//
// Usage: Class that handles the intermud <-> net communication
//
// All rights reserved. See license.doc for complete information.
//
// Copyright (C) 1993-1996 by the Hackerz of the Realm Of Magic
// Realm Of Magic is based on CircleMUD, Copyright (C) 1993.
// CircleMUD is based on DikuMUD, Copyright (C) 1990, 1991.
// Original intermud_client.c (C) 1995, 96 by Chris Austin (Stryker@Tempus)
// Original lpc protocol (C) by Mark Lewis (Nostradamus@Zebedee)
//                           by J.W. Armstrong (Alvin@Sushi)
// ============================================================================

#include <string>
#include <stdio.h>
#include <iostream.h>
#include <iomanip.h>

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "net.hh"
#include "log.hh"
#include "intermud.hh"

extern _log interlog;
extern void close_sockets();

void _net::Init(int port)
{
  struct sockaddr_in serv_addr;

  if ((file_descriptor = socket(PF_INET, SOCK_DGRAM, 0)) == -1) {
    interlog << "SYSERR: Can not open internet datagram socket!" << logit;
    exit(1);
  }
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port = htons(port);
  if (bind(file_descriptor, (struct sockaddr *) &serv_addr,
      sizeof(serv_addr)) == -1) {
    interlog << "SYSERR: Can not bind local address, port = " << port
             << logit;
    close_sockets();
    exit(1);
  }
  return;
}

string _net::ReceiveData(struct sockaddr_in &remote)
{
  char raw_data[MAX_NET_MESSAGE+500];
  
  int bytes;
  unsigned int addr_len;

  addr_len = sizeof(struct sockaddr_in);

  bytes = recvfrom(file_descriptor, raw_data, MAX_NET_MESSAGE, 0,
      (struct sockaddr *) &remote, &addr_len);

  raw_data[bytes] = '\0';
  if (bytes <= 0) {
    interlog << "ERROR: Zero bytes received on net socket" << logit;
    return "";
  } 
   
  in_data = raw_data;
  return in_data;
}

void _net::CloseFD()
{
  close(file_descriptor);
  return;
}

void _net::SetAddr(unsigned long int new_addr)
{
  ip_addr = new_addr;
  return;
}

void _net::SetPort(unsigned short int new_port)
{
  ip_port = new_port;
  return;
}

void _net::FlushData()
{
  struct sockaddr_in remote_address;

  remote_address.sin_family = AF_INET;
  remote_address.sin_addr.s_addr = ip_addr;
  remote_address.sin_port = ip_port;
  
  if (sendto(file_descriptor, out_data.c_str(), out_data.size(), 0,
      (struct sockaddr *) &remote_address, 
      sizeof(remote_address)) != out_data.size())
    interlog << "SYSERR: Errors while sending to network!" << logit;
}

void _net::AddToSet(fd_set &set)
{
  FD_SET(file_descriptor, &set);
  return;
}

bool _net::CheckSet(fd_set &set)
{
  if (FD_ISSET(file_descriptor, &set))
    return true;

  return false;
}

void _net::ClearSet(fd_set &set)
{
  FD_CLR(file_descriptor, &set);
  return;
}

int _net::Desc()
{
  return file_descriptor;
}

string _net::OutBuffer()
{
  return out_data;
}

_net& _net::operator<<(string s)
{
  this->out_data += s;
  return *this;
}

_net& _net::operator<<(int i)
{
  char num_buf[256];

  sprintf(num_buf, "%d", i);
  this->out_data += num_buf;
  return *this;
}

_net& _net::operator<<(unsigned int ui)
{
  char num_buf[256];

  sprintf(num_buf, "%u", ui);
  this->out_data += num_buf;
  return *this;
}

_net& _net::operator<<(long int li)
{
  char num_buf[256];

  sprintf(num_buf, "%ld", li);
  this->out_data += num_buf;
  return *this;
}

_net& _net::operator<<(unsigned long int uli)
{
  char num_buf[256];

  sprintf(num_buf, "%lu", uli);
  this->out_data += num_buf;
  return *this;
}

_net& _net::operator<<(short int si)
{
  return (*this) << (int)si;
}

_net& _net::operator<<(unsigned short int usi)
{
  return (*this) << (unsigned int)usi;
}

_net& _net::operator<<(_net& (*pf)(_net&))
{
  return (*pf)(*this);
}

_net& sendit(_net& flush_net)
{
  flush_net.FlushData();
  flush_net.out_data = "";
  flush_net.ip_addr = 0;
  flush_net.ip_port = 0;
  return flush_net;
}

